/*
 * events.c - Linux IEEE-1394 Subsystem: Event Propagation Mechanism
 * Copyright (C)1999 Emanuel Pirker <epirker@edu.uni-klu.ac.at>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/slab.h>

#include "ieee1394_types.h"
#include "events.h"

static struct hpsb_event_handler *event_handler_list = NULL;

static struct hpsb_event event_queue[0];

spinlock_t event_lock = SPIN_LOCK_UNLOCKED;

void hpsb_dispatch_event(event_t eventcode, struct hpsb_host *host, 
			int data_i, unsigned long data_l)
{
    unsigned long flags;
    struct hpsb_event_handler *ptr;

    spin_lock_irqsave(&event_lock, flags);
  
    /* create structure */
    event_queue->event = eventcode;
    event_queue->host = host;
    event_queue->data_i = data_i;
    event_queue->data_l = data_l;

    /* propagate event information */
    for (ptr = event_handler_list ; ptr != NULL ; ptr = ptr->next) {
	if (ptr->mask & event_queue->event) 
	    ptr->event_handler(event_queue);
    };

    spin_unlock_irqrestore(&event_lock, flags);
}

int hpsb_reg_event_handler( void (* event_handler) (struct hpsb_event *),
			    event_t mask, evt_priority_t priority)
{
    struct hpsb_event_handler *new, *ptr, *ptr2;
    unsigned long flags;

    if (mask) {
	/* register event handler */
        new = (struct hpsb_event_handler *) 
	    kmalloc( sizeof(struct hpsb_event_handler), GFP_KERNEL);
	if (!new) return -ENOSPC;

        new -> next = NULL;
        new -> event_handler = event_handler;
	new -> mask = mask;

	spin_lock_irqsave(&event_lock, flags);

        if (event_handler_list==NULL) {
            event_handler_list = new;
        } else {
	    if (priority != EVT_PRIO_NORMAL)
		HPSB_PRINT(KERN_INFO, "unimplemented priority code on registering event handler");
	    for (ptr = event_handler_list; ptr->next != NULL; 
		 ptr = ptr->next) ;
	    ptr -> next = new;
	}

	spin_unlock_irqrestore(&event_lock, flags);

    } else {
	/* unregister handler */
        spin_lock_irqsave(&event_lock, flags);
	for (ptr2 = NULL, ptr = event_handler_list; ptr;
	     ptr2 = ptr, ptr = ptr->next)
	    if (ptr->event_handler == event_handler){
		if (ptr2)
		    ptr2->next = ptr->next;
		else
		    event_handler_list = ptr->next;
		kfree(ptr);
	    }
	spin_unlock_irqrestore(&event_lock, flags);
	
    };

    return 0;
}
