/*
**  ieee1394_events.h - Linux IEEE 1394 subsystem; event propagation
**  Copyright (C)1999 Emanuel Pirker <epirker@edu.uni-klu.ac.at>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
*/

#ifndef IEEE1394_EVENTS_H
#define IEEE1394_EVENTS_H

/* One bit per event; max 63 events, 0 reserved for NO_EVENT */
#define EVT_NO_EVENT                       0 
/* Phy layer related Events */
#define EVT_BUS_RESET_STARTED              1
#define EVT_BUS_RESET_COMPLETED           (1<<1)
#define EVT_CONFIG_TIMEOUT                (1<<2)
#define EVT_CABLE_POWER_FAIL              (1<<3)
#define EVT_PORT_STATUS_CHANGED           (1<<4)
#define EVT_PORT_SUSPEND_RESUME_FAULT     (1<<5)
#define EVT_PHY_PACKET                    (1<<6)
#define EVT_PING_RESPONSE                 (1<<7)
/* Link layer related events */
#define EVT_CYCLE_TOO_LONG                (1<<8)
#define EVT_CYCLE_START                   (1<<9)
#define EVT_CYCLE_LOST                    (1<<10)
#define EVT_DUPLICATE_CHANNEL             (1<<11)
#define EVT_BUS_OCCUPANCY_VIOLATION       (1<<12)
#define EVT_UNKNOWN_TRANSACTION_CODE      (1<<13)
#define EVT_HEADER_CRC_ERROR              (1<<14)
/* Host controller driver module related events */
#define EVT_HOST_ADAPTER_ERROR            (1<<15)
#define EVT_HOST_ADAPTER_HUNG             (1<<16)
#define EVT_HOST_ADAPTER_RESET            (1<<17)
/* Configuration/Management related events */
#define EVT_GAP_COUNT_CHANGE              (1<<18)
#define EVT_CYCLE_MASTER_ACTIVATED        (1<<19)
#define EVT_CSR_CHANGE                    (1<<20)
#define EVT_GUIDS_RESOLVED                (1<<21)
#define EVT_DEVICE_REMOVAL                (1<<22)
#define EVT_CHANNEL_ALLOCATED             (1<<23)
#define EVT_CHANNEL_FREED                 (1<<24)
#define EVT_BANDWIDTH_ALLOCATED           (1<<25)
#define EVT_BANDWIDTH_FREED               (1<<26)
#define EVT_UNEXPECTED_CHANNEL            (1<<27)

/* Event handler priorities */
typedef enum { EVT_PRIO_NONE, EVT_PRIO_NORMAL, EVT_PRIO_HIGH, 
	       EVT_PRIO_HIGHEST } evt_priority_t;

typedef __u64 event_t;

/* This structure stores an event. 
   event.event refers to the event code (see #defines above)
   event.host is a pointer to the host the event is related to.
   event.data_i and event_l are event-specific data fields. */
struct hpsb_event {
    event_t event;
    struct hpsb_host *host;
    int data_i;
    unsigned long data_l;
};

#ifdef __KERNEL__

/* This structure stores an event handler. It has a next pointer
   to allow usage in linked lists, the bitmask of events this
   handler should be called upon, the priority value and a pointer
   to the handler itself */
struct hpsb_event_handler {
    struct hpsb_event_handler *next;
    event_t mask;
    evt_priority_t prio;
    void (* event_handler) (struct hpsb_event *event);
};

/* Notify the subsystem of an event. */
void hpsb_dispatch_event(event_t eventcode, struct hpsb_host *host, 
			 int data_i, unsigned long data_l);

/* Register a handler for a given mask of events with given priority.
   Returns zero on success */
int hpsb_reg_event_handler( void (* event_handler) (struct hpsb_event *),
			    event_t mask, evt_priority_t priority);

#endif

#endif
