/*
 * raw1394.c - Linux 1394 Subsystem Raw I/O driver
 * Copyright (C)1999 Emanuel Pirker <epirker@edu.uni-klu.ac.at>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <asm/uaccess.h>

#include "ieee1394_types.h"
#include "hosts.h"
#include "ieee1394_core.h"
#include "ieee1394_transactions.h"
#include "events.h"
#include "raw1394.h"

static struct raw1394_port ports[MAX_RAW_PORTS];
static int num_of_ports;

void handle_event(struct hpsb_event *event)
{
    int i;
    struct raw1394_port *port = NULL; 

    HPSB_TRACE();

    /* first we search to which port the event belongs */
    for (i = 0; i<MAX_RAW_PORTS; i++) {
	if (event->host == ports[i].host) {
	    port = &ports[i];
	    break;
	};
    };

    /* no port found? well, obviously we did our job correctly, though */
    if (!port) return;

    memcpy(&port->event_buf, event, sizeof(struct hpsb_event));
	
    if (port->fasync_queue)
	kill_fasync(port->fasync_queue, SIGIO);


}

int raw_open(struct inode *inode, struct file *file)
{
    int portnum = MINOR(inode->i_rdev);
    
    MOD_INC_USE_COUNT;
    
    /* the portnum must be present... */
    if (portnum > num_of_ports) {
	MOD_DEC_USE_COUNT;
	return -ENXIO;
    }

    /* ... usable ... */
    if (ports[portnum].usage_status == lck_UNUSABLE) {
	MOD_DEC_USE_COUNT;
	return -ENXIO;
    }
    
    /* ... and free to use */
    if (ports[portnum].usage_status == lck_USED) {
	MOD_DEC_USE_COUNT;
	return -EBUSY;
    };

    file->private_data = (void *) &ports[portnum];

    return 0;
}

int raw_fasync(int fd, struct file *file, int on)
{
    struct raw1394_port *port = (struct raw1394_port *) file->private_data;

    return fasync_helper(fd, file, on, &port->fasync_queue);
}

int raw_release(struct inode *inode, struct file *file)
{
    int portnum = MINOR(inode->i_rdev);
 
    if (ports[portnum].usage_status == lck_USED) {
	ports[portnum].usage_status = lck_UNUSED;
    };

    raw_fasync(-1, file, 0);

    MOD_DEC_USE_COUNT;
    return 0;
}

unsigned int raw_poll(struct file *file, poll_table *pt)
{
    struct raw1394_port *port = (struct raw1394_port *) file->private_data;
    unsigned int mask = 0;

    if ( ( port -> req_num > 0 ) && ( ! port -> dirtybuf ) )
	mask = mask | POLLIN | POLLRDNORM;
    if ( port -> req_num == 0 )
	mask = mask | POLLOUT | POLLWRNORM;
    
    return mask;
}

ssize_t raw_read(struct file *file, char *buffer, size_t count, 
		 loff_t *offset) 
{
    struct raw1394_port *port = (struct raw1394_port *) file->private_data;
    struct RAW1394request_t *req = port -> req_buf;

    while ((port -> req_num == 0) || (port -> dirtybuf)) {
	/* no data here */
	if ( file -> f_flags & O_NONBLOCK)
	    return -EAGAIN;
	HPSB_TRACE();
	interruptible_sleep_on( &port -> rd_q );
	if (signal_pending(current))
	    return -ERESTARTSYS;
    }

    copy_to_user( buffer, req, count );

    port -> req_num = 0;
    wake_up_interruptible( &port -> wr_q );

    return count;
};

ssize_t raw_write(struct file *file, const char *buffer, size_t count, 
                  loff_t *offset)
{
    struct raw1394_port *port = (struct raw1394_port *) file->private_data;
    struct RAW1394request_t *req = port -> req_buf;
    u32 busmask;

    while (port -> req_num > 0) {
	/* no buffer space */
	if (file -> f_flags & O_NONBLOCK)
	    return -EAGAIN;
	HPSB_TRACE();
	interruptible_sleep_on( &port -> wr_q );
	if (signal_pending(current))
	    return -ERESTARTSYS;
    };
	
    port -> req_num = 1;
    port -> dirtybuf = 1;
    copy_from_user( port -> req_buf, buffer, count );

    busmask = ( port->host->csr.node_ids >> 16 ) & 0xFFC0;
    HPSB_PRINT(KERN_INFO,"busmask 0x%x",busmask);

    switch( req -> type ) {
    case RAW1394_QUADREAD:
	req -> responseCode = hpsb_read(port -> host, 
					req -> destinationID + busmask, 
					req -> destinationOffset, 
					&req -> quadletData, 4);
	break;
    case RAW1394_QUADWRITE:
	req -> responseCode = hpsb_write(port -> host, 
					 req -> destinationID + busmask, 
					 req -> destinationOffset, 
					 &req -> quadletData, 4);
	break;
    case RAW1394_BLOCKREAD:
	req -> responseCode = hpsb_read(port -> host, 
					req -> destinationID + busmask, 
					req -> destinationOffset, 
					(quadlet_t *)&req -> block, 
					req -> length);
	break;
    case RAW1394_BLOCKWRITE:
	req -> responseCode = hpsb_write(port -> host, 
					 req -> destinationID + busmask, 
					 req -> destinationOffset, 
					 (quadlet_t *)&req -> block, 
					 req -> length);
	break;
    case RAW1394_LOCK:
	HPSB_PRINT(KERN_ERR,"lock request not implemented.");
#if 0
	req -> responseCode = hpsb_lock(port -> host, 
					req -> destinationID + busmask, 
					req -> destinationOffset, 
					req -> extcode,
					&req -> block[0], &req -> block[2]);
#endif
	break;
    default:
	HPSB_TRACE();
	HPSB_PRINT(KERN_ERR,"unknown request code 0x%x in raw device driver",
		   req->type);
    };
    
    port -> dirtybuf = 0;
    wake_up_interruptible( &port -> rd_q );

    file->f_pos += count;
    return count;
};

static int raw_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		     unsigned long arg)
{
    struct raw1394_port *port = (struct raw1394_port *) file->private_data;
    int retval = 0;

    switch (cmd) {
    case RAW1394_GET_PHYID:
	put_user((port->host->node_id & 0x3F), (nodeid_t *) arg);
	break;
    case RAW1394_GET_IRMID:
	put_user(port->host->irm_id & 0x3F, (nodeid_t *) arg);
	break;
    case RAW1394_GET_BMGRID:
	put_user(port->host->busmgr_id, (nodeid_t *) arg);
	break;
    case RAW1394_GET_NODECOUNT:
	put_user(port->host->node_count, (unsigned long *) arg);
	break;
    case RAW1394_RESET_BUS:
	port->host->template->devctl(port->host, RESET_BUS, arg);
	break;
    case RAW1394_SET_CHANNEL:
	break;
    case RAW1394_GET_CHANNEL:
	break;
    case RAW1394_ENA_ISO:
	break;
    case RAW1394_SET_CYCLE_COUNTER:
	port->host->template->devctl(port->host, SET_CYCLE_COUNTER, arg);
	break;
    case RAW1394_GET_CYCLE_COUNTER:
	{ 
	    unsigned long cycle_count = 0;
	    cycle_count = port->host->template->devctl(port->host, 
						   GET_CYCLE_COUNTER, 0);
	    put_user(cycle_count, (unsigned long *) arg);
	}
	break;
    case RAW1394_ACT_CYCLE_MASTER:
	port->host->template->devctl(port->host, ACT_CYCLE_MASTER, arg);
	break;
    case RAW1394_CANCEL_REQUESTS:
	port->host->template->devctl(port->host, CANCEL_REQUESTS, 0);
	break;
    case RAW1394_GET_EVENT_BUFFER:
	copy_to_user( (void *) arg, 
		      &port->event_buf, sizeof(struct hpsb_event));
	port->event_buf.event = EVT_NO_EVENT;
    default:
	return -EINVAL;
    }

    return retval;
}

struct file_operations raw1394_ops = {
    NULL, 
    raw_read,
    raw_write,
    NULL, /* no readdir */
    raw_poll, 
    raw_ioctl, 
    NULL, /* no mmap */
    raw_open,
    NULL, /* no flush */
    raw_release,
    NULL, /* no fsync */
    raw_fasync, 
    NULL, /* no check_media_change */
    NULL, /* no revalidate */
    NULL  /* no lock */
};

int cleanup_raw1394(void)
{
    hpsb_reg_event_handler( handle_event, 0, EVT_PRIO_NONE );

    unregister_chrdev(RAW1394_MAJOR, RAW1394_DRIVER_NAME );
    return 0;
}

int init_raw1394(void)
{
    struct hpsb_host *hostlist[MAX_RAW_PORTS];
    int i;

    if (register_chrdev(RAW1394_MAJOR, RAW1394_DRIVER_NAME, &raw1394_ops)) {
	HPSB_PRINT(KERN_ERR, "allocation of char major number %d failed",
		   RAW1394_MAJOR);
	return -EBUSY;
    }

    /* initialize ports array */
    num_of_ports = hpsb_get_host_list(hostlist, MAX_RAW_PORTS);
    for (i = 0; i < num_of_ports; i++) {
	ports[i].host = hostlist[i];
	ports[i].event_buf.event = EVT_NO_EVENT;
	ports[i].usage_status = lck_UNUSED;
        init_waitqueue_head(&ports[i].rd_q);
        init_waitqueue_head(&ports[i].wr_q);
    };

    hpsb_reg_event_handler( handle_event, 0xFFFFFFFFFFULL, EVT_PRIO_NORMAL );

    HPSB_PRINT(KERN_INFO, "IEEE-1394 Raw I/O driver loaded with %d device(s)",
	       num_of_ports);
    return 0;
}

#ifdef MODULE

/* EXPORT_NO_SYMBOLS; */

MODULE_AUTHOR("Emanuel Pirker <epirker@edu.uni-klu.ac.at>");
MODULE_DESCRIPTION("Linux IEEE-1394 Subsystem Raw I/O module");
MODULE_SUPPORTED_DEVICE("raw1394");

void cleanup_module(void)
{
    cleanup_raw1394();
}

int init_module(void)
{
    init_raw1394();
    return 0;
}

#endif /* MODULE */
